import copy
from collections import deque
from numpy import argmin
from . import Makespan
from . import PrecedConstraints


class TS1(object):
    """
    : Tabu Search for the critical workload
    """ 

    def __init__(self, data, initial_state, overview, tabu_size, max_count, max_steps):
        self.initial_state = initial_state
        self.max_count = max_count
        self.max_steps = max_steps
        self.tabu_size = tabu_size
        self.data = data
        self.overview = overview
        self.cur_steps = None
        self.status = None
        self.current = None
        self.best = None
        self.count = None
        self.best_old = None
        self.ex_solutions = 0

    def _clear(self):
        self.cur_steps = 0
        self.count = 0
        self.tabu_list = deque(maxlen=self.tabu_size)
        self.current = self.initial_state
        self.best = self.initial_state
        self.best_old = self.initial_state
        self.status = 'OK'
        self.cur_steps = None
        self.status = None

    def score(self, state):
        self.ex_solutions += 1
        G = state
        _, _, makespan = Makespan.score(G, self.data, self.overview)
        return makespan

    def neighborhood(self):
        G = self.current
        temp_neighborhood = []
        cp, _, _ = Makespan.score(G, self.data, self.overview)

        blocks = []
        isin = []
        for x in cp:
            if x not in isin:
                temp = []
                ma = x[2]
                m = ma
                dif = 1
                while m == ma and dif == 1:
                    isin.append(x)
                    j = cp.index(x)
                    temp.append(x)
                    j_new = j + 1
                    if j_new <= len(cp) - 1:
                        dif = j_new - j
                        x = cp[j_new]
                        m = x[2]
                    else:
                        break
                blocks.append(temp)
            else:
                continue
        moves = [k for k in blocks if len(k) > 1]

        if len(moves) > 1:
            for i in range(len(moves)):
                temp = copy.copy(G)
                cached_idx = {entry: idx for idx, entry in enumerate(temp)}
                if i == 0:
                    a, b = moves[i][-2:]
                    ind_a = cached_idx[a]
                    ind_b = cached_idx[b]
                    if a[0] != b[0]:
                        temp[ind_a] = b
                        temp[ind_b] = a
                        temp_neighborhood.append(temp)
                    else:
                        continue
                elif i == len(moves) - 1:
                    a, b = moves[i][:2]
                    ind_a = cached_idx[a]
                    ind_b = cached_idx[b]
                    if a[0] != b[0]:
                        temp[ind_a] = b
                        temp[ind_b] = a
                        temp_neighborhood.append(temp)
                    else:
                        continue
                else:
                    if len(moves[i]) > 2:
                        for r in range(2):
                            if r == 0:
                                a, b = moves[i][:2]
                                ind_a = cached_idx[a]
                                ind_b = cached_idx[b]
                                if a[0] != b[0]:
                                    temp[ind_a] = b
                                    temp[ind_b] = a
                                    temp_neighborhood.append(temp)
                                else:
                                    pass
                            else:
                                cached_idx = {entry: idx for idx, entry in enumerate(temp)}
                                ind_a = cached_idx[a]
                                ind_b = cached_idx[b]
                                if a[0] != b[0]:
                                    temp[ind_a] = b
                                    temp[ind_b] = a
                                    temp_neighborhood.append(temp)
                                else:
                                    continue
                    else:
                        a, b = moves[i][:2]
                        ind_a = cached_idx[a]
                        ind_b = cached_idx[b]
                        if a[0] != b[0]:
                            temp[ind_a] = b
                            temp[ind_b] = a
                            temp_neighborhood.append(temp)
                        else:
                            continue

        neighborhood = []
        for neighb in temp_neighborhood:
            neighborhood.append(PrecedConstraints.preced_constraints(neighb))
        return neighborhood

    def best_neighbor(self, neighbors):
        return neighbors[argmin([self.score(x) for x in neighbors])]

    def run(self):
        self._clear()

        self.cur_steps = 0
        while self.cur_steps <= self.max_steps:

            self.cur_steps += 1
            neighbors = self.neighborhood()
            if not neighbors:
                return self.best, self.score(self.best), self.ex_solutions  # self.score(self.best)

            neighborhood_best = self.best_neighbor(neighbors)

            while True:
                if all([x in self.tabu_list for x in neighbors]):
                    self.status = "TERMINATING - NO SUITABLE NEIGHBORS"
                    return self.best, self.score(self.best), self.ex_solutions
                if neighborhood_best in self.tabu_list:
                    if self.score(neighborhood_best) < self.score(self.best):
                        self.tabu_list.append(neighborhood_best)
                        self.best = neighborhood_best
                        break
                    else:
                        neighbors.remove(neighborhood_best)
                        neighborhood_best = self.best_neighbor(neighbors)
                else:
                    self.tabu_list.append(neighborhood_best)
                    self.current = neighborhood_best
                    if self.score(self.current) < self.score(self.best):
                        self.best = self.current
                    break

            if self.best == self.best_old:
                self.count += 1
                self.best_old = self.best
            else:
                self.count = 0
                self.best_old = self.best
            if self.count > self.max_count:
                return self.best, self.score(self.best), self.ex_solutions

        return self.best, self.score(self.best), self.ex_solutions  # self.score(self.best), self.status
