import copy
import random
from collections import deque
import numpy as np
from numpy import argmin


class TS2():
    """
    : Tabu search for makespan
    """
    
    def __init__(self, max_mach, data, data2, initial_state, tabu_size, max_count, max_steps):
        self.initial_state = initial_state
        self.max_count = max_count
        self.max_steps = max_steps
        self.tabu_size = tabu_size
        self.cur_steps = None
        self.current = None
        self.best = None
        self.data2 = data2
        self.data = data
        self.max_mach = max_mach
        self.count = None
        self.best_old = None

    def _clear(self):
        self.cur_steps = 0
        self.count = 0
        self.tabu_list = deque(maxlen=self.tabu_size)
        self.current = self.initial_state
        self.best = self.initial_state
        self.best_old = self.initial_state

    def score(self, state):
        nodes = copy.copy(state)
        workload = []
        for i in range(1, self.max_mach+1):
            processes = [node for node in nodes if node[2] == i]
            total_proc_time = sum([self.data[i] for i in processes])
            workload.append(total_proc_time)
        maxworkload = np.max(workload)
        
        return maxworkload
    
    def neighborhood(self):
        nodes = self.current
        n = len(nodes)
        neighborhood = []
        pool = [i for i, x in enumerate(nodes) if len(self.data2[(x[0], x[1])]) > 1]
        for i in range(10):
            nodes_temp = copy.copy(nodes)
            ind_x, ind_y = random.sample(pool, 2)
            x = nodes_temp[ind_x]
            y = nodes_temp[ind_y]
            m1_new = random.choice(self.data2[(x[0], x[1])])
            m2_new = random.choice(self.data2[(y[0], y[1])])
            x = list(x)
            y = list(y)
            x[2] = m1_new
            y[2] = m2_new
            x = tuple(x)
            y = tuple(y)
            nodes_temp[ind_x] = x
            nodes_temp[ind_y] = y
            neighborhood.append(nodes_temp)
        
        return neighborhood
    
    def best_neighbor(self, neighbors):
        return neighbors[argmin([self.score(x) for x in neighbors])]

    def run(self):
        self._clear()
        
        self.cur_steps = 0
        while self.cur_steps <= self.max_steps:
            
            self.cur_steps += 1

            neighbors = self.neighborhood()
            if not neighbors:
                    return self.best, self.score(self.best), self.tabu_list
            
            neighborhood_best = self.best_neighbor(neighbors)

            while True:
                if all([x in self.tabu_list for x in neighbors]):
                    return self.best, self.score(self.best)
                if neighborhood_best in self.tabu_list:
                    if self.score(neighborhood_best) < self.score(self.best):
                        self.best = neighborhood_best
                        self.tabu_list.append(neighborhood_best)
                        break
                    else:
                        neighbors.remove(neighborhood_best)
                        neighborhood_best = self.best_neighbor(neighbors)
                else:
                    self.tabu_list.append(neighborhood_best)
                    self.current = neighborhood_best
                    if self.score(self.current) < self.score(self.best):
                        self.best = self.current
                    break

            if self.best == self.best_old:
                self.count += 1
                self.best_old = self.best
            else:
                self.count = 0
                self.best_old = self.best
            if self.count > self.max_count:
                return self.best, self.score(self.best), self.tabu_list

        return self.best, self.score(self.best), self.tabu_list
